NUM_OCTAVE_INPUT_PORTS = 2
NUM_OCTAVE_DIGITAL_INPUT_PORTS = 5
NUM_OCTAVE_OUTPUT_PORTS = 5
NUM_OPX_PLUS_INPUT_PORTS = 2
NUM_OPX_PLUS_OUTPUT_PORTS = 10
NUM_OPX_PLUS_DIGITAL_OUTPUT_PORTS = 10
NUM_LF_FEM_INPUT_PORTS = 2
NUM_LF_FEM_OUTPUT_PORTS = 8
NUM_LF_FEM_DIGITAL_OUTPUT_PORTS = 8
NUM_MW_FEM_INPUT_PORTS = 2
NUM_MW_FEM_OUTPUT_PORTS = 8
NUM_MW_FEM_DIGITAL_OUTPUT_PORTS = 8
